from bulb.contrib.auth.hashers import _hash_password
from bulb.contrib.auth.exceptions import *
from bulb.db.utils import make_uuid
from bulb.db import node_models
from bulb.db.base import gdbh
from django.template.loader import render_to_string
from django.core.mail import send_mail
from django.conf import settings
import importlib.util
import warnings
import datetime


BASE_DIR = settings.BASE_DIR


class FakeClass:
    pass


class GroupPermissionRelationship(node_models.Relationship):
    rel_type = "CAN"
    direction = "from"
    start = "self"
    target = "Permission"
    auto = False
    on_delete = "PROTECT"
    unique = False


class UserPermissionRelationship(node_models.Relationship):
    rel_type = "CAN"
    direction = "from"
    start = "self"
    target = "Permission"
    auto = False
    on_delete = "PROTECT"
    unique = False

    def get(self, direction="bi", returned="node", order_by=None, limit=None, skip=None, desc=False, distinct=False, only=None,
            only_user_perms=False):
        user_instance = self._self_node_instance
        if user_instance.is_active_user:
            permissions_list = []

            # Get user's permissions.
            user_permissions = super().get(direction="from", returned="node", order_by=order_by, limit=limit, skip=skip,
                                           desc=desc, distinct=distinct, only=only)

            if user_permissions is not None:
                for permission in user_permissions:
                    permissions_list.append(permission)

            if only_user_perms is False:
                # Get user's groups' permissions.
                groups = user_instance.groups.get(returned="node", direction="from")
                if groups is not None:
                    for group in groups:
                        group_permissions = group.permissions.get(direction="from", returned="node", order_by=order_by,
                                                                  limit=limit, skip=skip, desc=desc, distinct=distinct,
                                                                  only=only)

                        if group_permissions is not None:
                            for perm in group_permissions:
                                if perm not in permissions_list:
                                    permissions_list.append(perm)

            if permissions_list:
                return permissions_list

        return None


class GroupRelationship(node_models.Relationship):
    rel_type = "IS_IN"
    direction = "from"
    start = "self"
    target = "Group"
    auto = False
    on_delete = "PROTECT"
    unique = False


class RelatedUserRelationship(node_models.Relationship):
    rel_type = "IS_IN"
    direction = "to"
    start = "User"
    target = "self"
    auto = False
    on_delete = "PROTECT"
    unique = False


class SessionRelationsip(node_models.Relationship):
    rel_type = "IS_SESSION_OF"
    direction = "to"
    start = "Session"
    target = "self"
    auto = False
    on_delete = "CASCADE"
    unique = True


class Permission(node_models.Node):

    codename = node_models.Property(unique=True,
                                    required=True)

    description = node_models.Property(unique=True,
                                       required=True)

    def __str__(self):
        return f'<Permission object(codename="{self.codename}", description="{self.description}")>'

    def __repr__(self):
        return f'<Permission object(codename="{self.codename}", description="{self.description}")>'

    def __hash__(self):
        return hash(self.codename) + hash(self.description)

    @classmethod
    def get(cls, uuid=None, codename=None, order_by=None, limit=None, skip=None, desc=False, only=None):
        """
        This method allow the retrieving of Permission (or of one of its children classes) instances.


        :param uuid: The Universal Unique Identifier of a permission to get an unique permission instance.

        :param codename: The codename of a permission to get an unique permission instance.

        :param order_by: Must be the name of the property with which the returned datas will be sorted.
                         Examples : "datetime", "first_name", etc...

        :param limit: Must be an integer. This parameter defines the number of returned elements.

        :param skip: Must be an integer. This parameter defines the number of skipped elements. For example if
                     self.skip = 3, the 3 first returned elements will be skipped.

        :param desc: Must be a boolean. If it is False the elements will be returned in an increasing order, but it is
                     True, they will be returned in a descending order.

        :param only: Must be a list of field_names. If this parameter is filled, the return will not be Permission
                     instances, but a dict with "only" the mentioned fields.

        :return: If uuid is None, a list will be returned. Else it will be a unique instance.
        """
        property_statement = ""
        order_by_statement = ""
        limit_statement = ""
        skip_statement = ""
        desc_statement = ""

        # Build the property_statement.
        if uuid is not None and codename is not None:
            property_statement = "{uuid:'%s', codename:'%s'}" % (uuid, codename)

        elif uuid is not None:
            property_statement = "{uuid:'%s'}" % uuid

        elif codename is not None:
            property_statement = "{codename:'%s'}" % codename

        # Build the match_statement.
        cyher_labels = node_models.DatabaseNode.format_labels_to_cypher(cls._get_labels())
        match_statement = f"MATCH (p:{cyher_labels} {property_statement})"

        # Build the with_statement.
        with_statement = "WITH p"

        # Build order_by statements.
        if order_by is not None:
            order_by_statement = f"ORDER BY p.{order_by}"

        # Build return_statement statements.
        if not only:
            return_statement = "RETURN (p)"

        else:
            only_statement_list = []

            for element in only:
                only_statement_list.append(f"p.{element}")

            only_statement = ", ".join(only_statement_list)

            return_statement = f"RETURN {only_statement}"

        # Build limit_statement.
        if limit is not None:
            if not isinstance(limit, str) and not isinstance(limit, int):
                raise BULBGetUserError(
                    f"The 'limit' parameter of the get() method of {cls.__name__} must be a string or an integer.")

            else:
                limit_statement = f"LIMIT {limit}"

        # Build skip_statement and add its required variable.
        if skip is not None:
            if not isinstance(skip, str) and not isinstance(skip, int):
                raise BULBGetUserError(
                    f"The 'skip' parameter of the get() method of {cls.__name__} must be a string or an integer.")

            else:
                skip_statement = f"SKIP {skip}"

        # Build desc_statement.
        if not isinstance(desc, bool):
            raise BULBGetUserError(
                f"The 'desc' parameter of the get() method of {cls.__name__} must be a boolean.")

        else:
            if desc is True:
                desc_statement = "DESC"

        request_statement = """
            %s
            %s
            %s
            %s
            %s
            %s
            %s
            """ % (match_statement,
                   with_statement,
                   order_by_statement,
                   desc_statement,
                   limit_statement,
                   skip_statement,
                   return_statement)

        response = gdbh.r_transaction(request_statement)

        if response:
            if only is None:
                fake_instances_list = []

                for permission_object in response:
                    fake_instances_list.append(cls.build_fake_instance(permission_object["p"],
                                                                       forced_fake_instance_class=cls))

                if uuid is not None or codename is not None:
                    return fake_instances_list[0]
                else:
                    return fake_instances_list

            else:
                return response

        else:
            return None


def get_permission_node_model():
    if settings.BULB_PERMISSION_NODE_MODEL_FILE == "bulb.contrib.auth.node_models":
        return Permission

    else:
        overloaded_permission = None
        overloaded_permission_module = importlib.import_module(settings.BULB_PERMISSION_NODE_MODEL_FILE)

        try:
            overloaded_permission = overloaded_permission_module.Permission

        except AttributeError:
            warnings.warn(
                f"You have defined BULB_PERMISSION_NODE_MODEL_FILE = '{settings.BULB_PERMISSION_NODE_MODEL_FILE}' but no Permission node model was found in it. So the native Permission node model will be used.",
                BULBPermissionWarning)

        else:
            return overloaded_permission
        return Permission


class Group(node_models.Node):

    name = node_models.Property(unique=True,
                                required=True)

    permissions = GroupPermissionRelationship()

    users = RelatedUserRelationship()

    def __str__(self):
        return f'<Group object(name="{self.name}", uuid="{self.uuid}")>'

    def __repr__(self):
        return f'<Group object(name="{self.name}", uuid="{self.uuid}")>'

    @classmethod
    def get(cls, uuid=None, name=None, order_by=None, limit=None, skip=None, desc=False, only=None):
        """
        This method allow the retrieving of Group (or of one of its children classes) instances.

        :param uuid: The Universal Unique Identifier of a group to get an unique group instance.

        :param name: The name of a group to get an unique group instance.

        :param order_by: Must be the name of the property with which the returned datas will be sorted.
                         Examples : "datetime", "first_name", etc...

        :param limit: Must be an integer. This parameter defines the number of returned elements.

        :param skip: Must be an integer. This parameter defines the number of skipped elements. For example if
                     self.skip = 3, the 3 first returned elements will be skipped.

        :param desc: Must be a boolean. If it is False the elements will be returned in an increasing order, but it is
                     True, they will be returned in a descending order.

        :param only: Must be a list of field_names. If this parameter is filled, the return will not be Group instances,
                     but a dict with "only" the mentioned fields.

        :return: If uuid is None, a list will be returned. Else it will be a unique instance.
        """
        property_statement = ""
        order_by_statement = ""
        limit_statement = ""
        skip_statement = ""
        desc_statement = ""

        # Build the property_statement.
        if uuid is not None and name is not None:
            property_statement = "{uuid:'%s', name:'%s'}" % (uuid, name)

        elif uuid is not None:
            property_statement = "{uuid:'%s'}" % uuid

        elif name is not None:
            property_statement = "{name:'%s'}" % name

        # Build the match_statement.
        cyher_labels = node_models.DatabaseNode.format_labels_to_cypher(cls._get_labels())
        match_statement = f"MATCH (g:{cyher_labels} {property_statement})"

        # Build the with_statement.
        with_statement = "WITH g"

        # Build order_by statements.
        if order_by is not None:
            order_by_statement = f"ORDER BY g.{order_by}"

        # Build return_statement statements.
        if not only:
            return_statement = "RETURN (g)"

        else:
            only_statement_list = []

            for element in only:
                only_statement_list.append(f"g.{element}")

            only_statement = ", ".join(only_statement_list)

            return_statement = f"RETURN {only_statement}"

        # Build limit_statement.
        if limit is not None:
            if not isinstance(limit, str) and not isinstance(limit, int):
                raise BULBGetUserError(
                    f"The 'limit' parameter of the get() method of {cls.__name__} must be a string or an integer.")

            else:
                limit_statement = f"LIMIT {limit}"

        # Build skip_statement and add its required variable.
        if skip is not None:
            if not isinstance(skip, str) and not isinstance(skip, int):
                raise BULBGetUserError(
                    f"The 'skip' parameter of the get() method of {cls.__name__} must be a string or an integer.")

            else:
                skip_statement = f"SKIP {skip}"

        # Build desc_statement.
        if not isinstance(desc, bool):
            raise BULBGetUserError(
                f"The 'desc' parameter of the get() method of {cls.__name__} must be a boolean.")

        else:
            if desc is True:
                desc_statement = "DESC"

        request_statement = """
            %s
            %s
            %s
            %s
            %s
            %s
            %s
            """ % (match_statement,
                   with_statement,
                   order_by_statement,
                   desc_statement,
                   limit_statement,
                   skip_statement,
                   return_statement)

        response = gdbh.r_transaction(request_statement)

        if response:
            if only is None:
                fake_instances_list = []

                for group_object in response:
                    fake_instances_list.append(cls.build_fake_instance(group_object["g"],
                                                                       forced_fake_instance_class=cls))

                if uuid is not None or name is not None:
                    return fake_instances_list[0]
                else:
                    return fake_instances_list

            else:
                return response

        else:
            return None


def get_group_node_model():
    if settings.BULB_GROUP_NODE_MODEL_FILE == "bulb.contrib.auth.node_models":
        return Group

    else:
        overloaded_group = None
        overloaded_group_module = importlib.import_module(settings.BULB_GROUP_NODE_MODEL_FILE)

        try:
            overloaded_group = overloaded_group_module.Group

        except AttributeError:
            warnings.warn(
                f"You have defined BULB_GROUP_NODE_MODEL_FILE = '{settings.BULB_GROUP_NODE_MODEL_FILE}' but no Group node model was found in it. So the native Group node model will be used.",
                BULBGroupWarning)

        else:
            return overloaded_group
        return Group


class EmptyRelationship(node_models.Relationship):
    rel_type = ""


class AnonymousUser:
    is_super_user = False
    is_staff_user = False
    is_active_user = False
    uuid = None
    first_name = None
    last_name = None
    email = None
    password = None
    registration_datetime = None
    session = EmptyRelationship()
    groups = EmptyRelationship()
    permissions = EmptyRelationship()

    def __str__(self):
        return 'AnonymousUser'

    def __repr__(self):
        return 'AnonymousUser'

    @classmethod
    def create(cls, **extrafields):
        raise NotImplementedError("bulb doesn't provide a DB representation for AnonymousUser.")

    @classmethod
    def create_super_user(cls, **extrafields):
        raise NotImplementedError("bulb doesn't provide a DB representation for AnonymousUser.")

    @classmethod
    def get(cls, uuid_or_email):
        return cls()

    def update(self, user_property, new_user_property_value):
        raise NotImplementedError("bulb doesn't provide a DB representation for AnonymousUser.")

    def delete(self):
        raise NotImplementedError("bulb doesn't provide a DB representation for AnonymousUser.")

    def set_password(self, new_password):
        raise NotImplementedError("bulb doesn't provide a DB representation for AnonymousUser.")

    def has_perm(self, permission_code_name):
        return False


def get_anonymoususer_node_model():
    if settings.BULB_ANONYMOUSUSER_NODE_MODEL_FILE == "bulb.contrib.auth.node_models":
        return AnonymousUser
    else:
        overloaded_anonymoususer = None
        overloaded_anonymoususer_module = importlib.import_module(settings.BULB_ANONYMOUSUSER_NODE_MODEL_FILE)
        try:
            overloaded_anonymoususer = overloaded_anonymoususer_module.AnonymousUser
        except AttributeError:
            warnings.warn(
                f"You have defined BULB_ANONYMOUSUSER_NODE_MODEL_FILE = '{settings.BULB_ANONYMOUSUSER_NODE_MODEL_FILE}' but no AnonymousUser node model was found in it. So the native AnonymousUser node model will be used.",
                BULBAnonymousUserWarning)
        else:
            return overloaded_anonymoususer
        return AnonymousUser


class User(node_models.Node):

    # Default groups
    is_super_user = node_models.Property(default=False)

    is_staff_user = node_models.Property(default=False)

    is_active_user = node_models.Property(default=True)

    first_name = node_models.Property(required=True)

    last_name = node_models.Property(required=True)

    email = node_models.Property(required=True,
                                 unique=True)

    password = node_models.Property(required=True)

    registration_datetime = node_models.Property(default=datetime.datetime.now)

    permissions = UserPermissionRelationship()

    groups = GroupRelationship()

    session = SessionRelationsip()

    email_confirmation_key = None

    email_is_confirmed = None

    if settings.BULB_REGISTRATION_USE_EMAIL_CONFIRMATION:
        email_confirmation_key = node_models.Property()

        email_is_confirmed = node_models.Property(default=False)

    def __str__(self):
        return f'<User object(first_name="{self.first_name}", last_name="{self.last_name}", uuid="{self.uuid}")>'

    def __repr__(self):
        return f'<User object(first_name="{self.first_name}", last_name="{self.last_name}", uuid="{self.uuid}")>'

    @classmethod
    def create(cls, **extrafields):

        extrafields.setdefault('is_super_user', False)
        extrafields.setdefault('is_staff_user', False)
        extrafields.setdefault('is_active_user', True)

        extrafields['password'] = _hash_password(extrafields['password'])

        new_user = cls(**extrafields)

        if settings.BULB_REGISTRATION_USE_EMAIL_CONFIRMATION:
            # Generate the email confirmation key.
            email_confirmation_key = make_uuid()

            # Apply it to the new User instance.
            new_user.update('email_confirmation_key', email_confirmation_key)

            target_email = extrafields[settings.BULB_USER_EMAIL_PROPERTY_NAME]
            confirmation_url = settings.BULB_CONFIRMATION_VIEW_PATH + email_confirmation_key

            send_mail(
                from_email=settings.BULB_EMAIL_SENDER_NAME,
                subject=settings.BULB_EMAIL_SUBJECT,
                recipient_list=[f"{target_email}"],
                message=settings.BULB_EMAIL_DEFAULT_MESSAGE + confirmation_url,
                html_message=render_to_string(settings.BULB_EMAIL_TEMPLATE_PATH,
                                              {"confirmation_url": confirmation_url}))

        return new_user

    @classmethod
    def create_super_user(cls, **extrafields):
        extrafields.setdefault('is_super_user', True)
        extrafields.setdefault('is_staff_user', True)
        extrafields.setdefault('is_active_user', True)

        extrafields['password'] = _hash_password(extrafields['password'])

        is_super_user = extrafields.get('is_super_user')
        is_staff_user = extrafields.get('is_staff_user')
        if not is_super_user:
            raise ValueError('A SuperUser must have "is_super_user = True".')

        elif not is_staff_user:
            raise ValueError('A SuperUser must have "is_staff_user = True".')

        else:
            new_super_user = User(**extrafields)

            return new_super_user

    @classmethod
    def get(cls, uuid=None, email=None, email_confirmation_key=None, order_by=None, limit=None, skip=None, desc=False, only=None):
        """
        This method allow the retrieving of User (or of one of its children classes) instances.

        :param uuid: The Universal Unique Identifier of a user to get an unique user instance.

        :param email: The email of a user to get an unique user instance.

        :param order_by: Must be the name of the property with which the returned datas will be sorted.
                         Examples : "datetime", "first_name", etc...

        :param limit: Must be an integer. This parameter defines the number of returned elements.

        :param skip: Must be an integer. This parameter defines the number of skipped elements. For example if
                     self.skip = 3, the 3 first returned elements will be skipped.

        :param desc: Must be a boolean. If it is False the elements will be returned in an increasing order, but it is
                     True, they will be returned in a descending order.

        :param only: Must be a list of field_names. If this parameter is filled, the return will not be User instances,
                     but a dict with "only" the mentioned fields.

        :return: If uuid is None, a list will be returned. Else it will be a unique instance.
        """
        property_statement = ""
        order_by_statement = ""
        limit_statement = ""
        skip_statement = ""
        desc_statement = ""

        # Build the property_statement.
        if uuid is not None and email is not None:
            property_statement = "{uuid:'%s', email:'%s'}" % (uuid, email)

        elif uuid is not None:
            property_statement = "{uuid:'%s'}" % uuid

        elif email is not None:
            property_statement = "{email:'%s'}" % email

        elif email_confirmation_key is not None:
            property_statement = "{email_confirmation_key:'%s'}" % email_confirmation_key

        # Build the match_statement.
        cyher_labels = node_models.DatabaseNode.format_labels_to_cypher(cls._get_labels())
        match_statement = f"MATCH (u:{cyher_labels} {property_statement})"

        # Build the with_statement.
        with_statement = "WITH u"

        # Build order_by statements.
        if order_by is not None:
            order_by_statement = f"ORDER BY u.{order_by}"

        # Build return_statement statements.
        if not only:
            return_statement = "RETURN (u)"

        else:
            only_statement_list = []

            for element in only:
                only_statement_list.append(f"u.{element}")

            only_statement = ", ".join(only_statement_list)

            return_statement = f"RETURN {only_statement}"

        # Build limit_statement.
        if limit is not None:
            if not isinstance(limit, str) and not isinstance(limit, int):
                raise BULBGetUserError(
                    f"The 'limit' parameter of the get() method of {cls.__name__} must be a string or an integer.")

            else:
                limit_statement = f"LIMIT {limit}"

        # Build skip_statement and add its required variable.
        if skip is not None:
            if not isinstance(skip, str) and not isinstance(skip, int):
                raise BULBGetUserError(
                    f"The 'skip' parameter of the get() method of {cls.__name__} must be a string or an integer.")

            else:
                skip_statement = f"SKIP {skip}"

        # Build desc_statement.
        if not isinstance(desc, bool):
            raise BULBGetUserError(
                f"The 'desc' parameter of the get() method of {cls.__name__} must be a boolean.")

        else:
            if desc is True:
                desc_statement = "DESC"

        request_statement = """
               %s
               %s
               %s
               %s
               %s
               %s
               %s
               """ % (match_statement,
                      with_statement,
                      order_by_statement,
                      desc_statement,
                      limit_statement,
                      skip_statement,
                      return_statement)

        response = gdbh.r_transaction(request_statement)

        if response:
            if only is None:

                fake_instances_list = []

                for user_object in response:
                    fake_instances_list.append(cls.build_fake_instance(user_object["u"],
                                                                       forced_fake_instance_class=cls))

                if uuid is not None or email is not None or email_confirmation_key is not None:
                    return fake_instances_list[0]

                else:
                    return fake_instances_list

            else:
                return response

        else:
            if uuid is not None or email is not None or email_confirmation_key is not None:
                if settings.BULB_ANONYMOUSUSER_NODE_MODEL_FILE:
                    AnonymousUser_node_model = get_anonymoususer_node_model()
                    return AnonymousUser_node_model()

            else:
                return None

    def set_password(self, new_password):
        self.update("password", _hash_password(new_password))

    def has_perm(self, permission_code_name):
        if self.is_super_user:
            return True

        if not self.is_active_user:
            return False

        else:
            permission_to_test = Permission.get(codename=permission_code_name)
            all_user_permissions = self.permissions.get()

            if all_user_permissions is not None:
                if permission_to_test in all_user_permissions:
                    return True
            return False


def get_user_node_model():
    if settings.BULB_USER_NODE_MODEL_FILE == "bulb.contrib.auth.node_models":
        return User

    else:
        overloaded_user = None
        overloaded_user_module = importlib.import_module(settings.BULB_USER_NODE_MODEL_FILE)

        try:
            overloaded_user = overloaded_user_module.User

        except AttributeError:
            warnings.warn(
                f"You have defined BULB_USER_NODE_MODEL_FILE = '{settings.BULB_USER_NODE_MODEL_FILE}' but no User node model was found in it. So the native User node model will be used.",
                BULBUserWarning)

        else:
            return overloaded_user
        return User



User_preview_fields = {"1": "uuid",
                       "2": "last_name",
                       "3": "first_name",
                       "4": "email",
                       "5": "is_super_user",
                       "6": "is_staff_user",
                       "7": "is_active_user",
                       "8": "is_subscribed_user"}

User_fields_infos = {"is_super_user": {"type": "checkbox",
                                       "label": "Super administrateur",
                                       "description": "Un super administrateur n'est soumis à aucune restriction. Il possède toutes les permissions."},

                     "is_staff_user": {"type": "checkbox",
                                       "label": "Administrateur",
                                       "description": "Un administrateur peut accéder aux différentes pages de l'interface d'administration. Cependant, cela ne lui donne aucunement le droit de créer, modifier ou supprimer quoi que ce soit."},

                     "is_active_user": {"type": "checkbox",
                                        "label": "Actif",
                                        "description": "Un compte utilisateur actif peut permettre de s'authentifier sur le site. Un utilisateur inactif est par exemple un compte utilisateur banni, qui ne pourra donc plus se connecter au site."},

                     "uuid": {"type": "locked",
                              "label": "UUID",
                              "description": "C'est un identifiant unique et universel permettant une différenciation stricte entre chacun des noeuds."},

                     "first_name": {"type": "text",
                                    "label": "Prénom"},

                     "last_name": {"type": "text",
                                   "label": "NOM"},

                     "email": {"type": "text",
                               "label": "Adresse email"},

                     "password": {"type": "password",
                                  "label": "Mot de passe"},

                     "registration_datetime": {"type": "datetime",
                                               "label": "Date et heure d'inscription"},

                     "session": {"type": "relationship",
                                 "label": "Session",
                                 "rel": {
                                     "unique": True,
                                     "related_node_model_name": "Session",
                                     "choices_render": ["session_key"],
                                 }},

                     "permissions": {"type": "relationship",
                                     "label": "Permissions",
                                     "rel": {
                                         "related_node_model_name": "Permission",
                                         "choices_render": ["codename"],
                                     }},

                     "groups": {"type": "relationship",
                                "label": "Groupes",
                                "rel": {
                                    "related_node_model_name": "Group",
                                    "choices_render": ["name"],
                                }}
                     }

Permission_preview_fields = {"1": "uuid",
                             "2": "codename",
                             "3": "description"}

Permission_fields_infos = {"uuid": {"type": "locked",
                                    "label": "UUID",
                                    "description": "C'est un identifiant unique et universel permettant une différenciation stricte entre chacun des noeuds."},
                           "codename": {"type": "text",
                                        "label": "Nom de code"},

                           "description": {"type": "text",
                                           "label": "Actif"},

                           }

Group_preview_fields = {"1": "uuid",
                        "2": "name"}

Group_fields_infos = {"uuid": {"type": "locked",
                                    "label": "UUID",
                                    "description": "C'est un identifiant unique et universel permettant une différenciation stricte entre chacun des noeuds."},
                      "name": {"type": "text",
                               "label": "Nom"},

                      "users": {"type": "relationship",
                                "label": "Utilisateurs",
                                "rel": {
                                    "related_node_model_name": "User",
                                    "choices_render": ["first_name", "last_name"],
                                }},

                      "permissions": {"type": "relationship",
                                      "label": "Permissions",
                                      "rel": {
                                          "related_node_model_name": "Permission",
                                          "choices_render": ["codename"],
                                      }},
                     }


from bulb.contrib.auth.node_models import Permission
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    args = ''
    help = """
            Initialize the bulb native permissions.
            """

    def handle(self, *args, **options):

        # Initial permissions :
        permissions_dict = {
            "create": "Permission.create(codename='create', description='The user can create everything.')",
            "view": "Permission.create(codename='view', description='The user can view everything.')",
            "update": "Permission.create(codename='update', description='The user can update everything.')",
            "delete": "Permission.create(codename='delete', description='The user can delete everything.')",

            "create_user": "Permission.create(codename='create_user', description='The user can create an User instance.')",
            "view_user": "Permission.create(codename='view_user', description='The user can view an User instance.')",
            "update_user": "Permission.create(codename='update_user', description='The user can update an User instance.')",
            "delete_user": "Permission.create(codename='delete_user', description='The user can delete an User instance.')",

            "create_group": "Permission.create(codename='create_group', description='The user can create a Group instance.')",
            "view_group": "Permission.create(codename='view_group', description='The user can view a Group instance.')",
            "update_group": "Permission.create(codename='update_group', description='The user can update a Group instance.')",
            "delete_group": "Permission.create(codename='delete_group', description='The user can delete a Group instance.')",

            "create_permission": "Permission.create(codename='create_permission', description='The user can create a Permission instance.')",
            "view_permission": "Permission.create(codename='view_permission', description='The user can view a Permission instance.')",
            "update_permission": "Permission.create(codename='update_permission', description='The user can update a Permission instance.')",
            "delete_permission": "Permission.create(codename='delete_permission', description='The user can delete a Permission instance.')",
        }

        print("\n-----------------------------------")
        print("     Start of initialization")
        print("-----------------------------------")

        for permission_name, permission in permissions_dict.items():
            p = Permission.get(codename=permission_name)
            if p is None:
                eval(permission)
                print(f"✔   '{permission_name}' has been created in the database.")

            else:
                print(f"❌   '{permission_name}' was already created.")

        print("--------------------------------")
        print("     End of initialization")
        print("--------------------------------\n")





from bulb.sftp_and_cdn.exceptions import BULBStaticfilesError
from bulb.utils import get_folders_paths_list
from bulb.sftp_and_cdn.sftp import SFTP
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    args = ''
    help = """
            Push source staticfiles folder to the SFTP server.
            """

    def add_arguments(self, parser):
        parser.add_argument("-r", "--raw", action="store_true")
        parser.add_argument("-b", "--bundled", action="store_true")

    def handle(self, *args, **kwargs):

        # Get the staticfiles folder path
        raw_staticfiles_folder_path = get_folders_paths_list("staticfiles")

        if raw_staticfiles_folder_path is not None:
            bundled_staticfiles_folder_path = get_folders_paths_list("bundled_staticfiles")

            if kwargs["raw"]:
                SFTP.push_src_staticfiles(src_type="raw",
                                          local_staticfiles_folder_path=raw_staticfiles_folder_path)

            if kwargs["bundled"]:
                if bundled_staticfiles_folder_path is not None:
                    SFTP.push_src_staticfiles(src_type="bundled",
                                              local_staticfiles_folder_path=bundled_staticfiles_folder_path)

                else:
                    raise BULBStaticfilesError(
                        "The bundled_staticfiles folder was not found. Please execute the 'collectstatic' and 'bundlestatic' commands before pushing it to your sftp server.")

            if not kwargs["raw"] and not kwargs["bundled"]:
                SFTP.push_src_staticfiles(src_type="raw",
                                          local_staticfiles_folder_path=raw_staticfiles_folder_path)

                if bundled_staticfiles_folder_path is not None:
                    SFTP.push_src_staticfiles(src_type="bundled",
                                              local_staticfiles_folder_path=bundled_staticfiles_folder_path)
                else:
                    raise BULBStaticfilesError(
                        "The bundled_staticfiles folder was not found. Please execute the 'collectstatic' and 'bundlestatic' commands before pushing it to your sftp server.")
        else:
            raise BULBStaticfilesError(
                "The staticfiles folder was not found. Please execute the 'collectstatic' command before pushing it to your sftp server.")



from django.template.defaultfilters import register


@register.filter()
def get(instance, parameter):
    value = None
    try:
        if isinstance(instance, dict):
            value = instance[parameter]

        else:
            value = eval(f"{instance}.{parameter}")

    except:
        value = instance.__dict__[parameter]

    finally:
        return value



Session_preview_fields = {"1": "uuid",
                          "2": "session_key",
                          "3": "expire_date"}

Session_fields_infos = {"uuid": {"type": "locked",
                                 "label": "UUID",
                                 "description": "C'est un identifiant unique et universel permettant une différenciation stricte entre chacun des noeuds."},

                        "session_key": {"type": "locked",
                                        "label": "Clé de session",
                                        "description": "Un clé de session est générée automatiquement pour chaque utilisateur qui se connecte. Elle permet d'identifier sa session et de garantir l'identité de l'utilisateur."},

                        "session_data": {"type": "locked",
                                         "label": "Données de session",
                                         "description": "Ces données chiffrées en base64 stockent l'UUID de l'utilisateur relié à cette session."},

                        "expire_date": {"type": "datetime",
                                        "label": "Date et heure d'expiration",
                                        "description": "Ces données chiffrées en base64 stockent l'UUID de l'utilisateur ainsi que la date et l'heure d'expiration de la session."},

                        "related_user": {"type": "relationship",
                                         "label": "Utilisateur",
                                         "rel": {
                                            "unique": True,
                                            "related_node_model_name": "User",
                                            "choices_render": ["first_name", "last_name"],
                                         }},
                        }


from bulb.sftp_and_cdn.exceptions import BULBStaticfilesError
from django.core.management.base import BaseCommand
from django.conf import settings
import subprocess


BASE_DIR = settings.BASE_DIR


class Command(BaseCommand):
    args = ''
    help = """
            Collect, clear, bundle and push the staticfiles folder to the SFTP server and purge the CDN if there is one.
            """

    def handle(self, *args, **options):

        if not settings.DEBUG:
            """
            If the command is executed when DEBUG = False, errors can occur from the bulb templatetags.
            Indeed, if DEBUG = False, if {% static_raw_src %}, {% static_bundled_src %} are used, staticfiles will be
            recovered from the SFTP : this will cause that the 'collectstatic command will collect staticfiles from the 
            old files on the SFTP, and not from the local new files.
            But also in certain cases, bundled staticfiles will be used : this will cause that the new bundle files will 
            be build from the old bundle files.

            See : TODO : add the related documentation page url.
            """
            raise BULBStaticfilesError("The DEBUG variable must be set on 'True' to handle new staticfiles.")

        else:
            # Collect staticfiles.
            print("\n--------------------------")
            print("-- COLLECT STATIC FILES --")
            print("--------------------------\n")

            subprocess.call("python manage.py collectstatic --clear", shell=True)

            # Clear old source staticfiles and purge them on the CDN if there is one.
            subprocess.call("python manage.py clearstatic", shell=True)

            if settings.BULB_SFTP_SRC_STATICFILES_MODE == "raw" or settings.BULB_SFTP_SRC_STATICFILES_MODE == "both":

                # Push new source staticfiles on the SFTP.
                subprocess.call("python manage.py pushstatic --raw", shell=True)

            if settings.BULB_SFTP_SRC_STATICFILES_MODE == "bundled" or settings.BULB_SFTP_SRC_STATICFILES_MODE == "both":

                # Generate the bundled_staticfiles folder which contains all the bundle of the project staticfiles.
                subprocess.call("python manage.py bundlestatic", shell=True)

                # Push new source staticfiles on the SFTP.
                subprocess.call("python manage.py pushstatic --bundled", shell=True)



import os


class BULB_BASE_DIRError(Exception):
    pass


def set_bulb_settings_on(root_settings):
    """
    This function overwrite many settings variables from the project's settings.py file, in order
    to add the bulb modules to the project but also to make compatible the django project with bulb.
    :param root_settings: The locals() dict of the project settings.py file
    """
    # Set the BASE_DIR variable as environment variable.
    try:
        BASE_DIR = root_settings['BASE_DIR']

    except KeyError:
        raise BULB_BASE_DIRError("You must define the default BASE_DIR variable into the settings.py file of your project.")

    else:
        os.environ["BASE_DIR"] = BASE_DIR

    # Add the bulb module.
    try:
        INSTALLED_APPS = root_settings['INSTALLED_APPS']

    except KeyError:
        root_settings['INSTALLED_APPS'] = ['bulb', ]

    else:
        INSTALLED_APPS.insert(0, 'bulb')

    # Remove the native auth module and set the bulb one.
    try:
        INSTALLED_APPS = root_settings['INSTALLED_APPS']

    except KeyError:
        root_settings['INSTALLED_APPS'] = ['bulb.contrib.auth', ]

    else:
        if 'django.contrib.auth' in INSTALLED_APPS:
            INSTALLED_APPS.remove('django.contrib.auth')
            INSTALLED_APPS.insert(0, 'bulb.contrib.auth')
        else:
            pass
            INSTALLED_APPS.insert(0, 'bulb.contrib.auth')

    # Remove the native admin module and set the bulb one.
    try:
        INSTALLED_APPS = root_settings['INSTALLED_APPS']

    except KeyError:
        root_settings['INSTALLED_APPS'] = ['bulb.contrib.admin', ]

    else:
        if 'django.contrib.admin' in INSTALLED_APPS:
            INSTALLED_APPS.remove('django.contrib.admin')
            INSTALLED_APPS.insert(0, 'bulb.contrib.admin')
        else:
            INSTALLED_APPS.insert(0, 'bulb.contrib.admin')

    # Remove the native contenttypes module and set the bulb one.
    try:
        INSTALLED_APPS = root_settings['INSTALLED_APPS']

    except KeyError:
        pass
        # root_settings['INSTALLED_APPS'] = ['bulb.contrib.contenttypes', ]

    else:
        if 'django.contrib.contenttypes' in INSTALLED_APPS:
            INSTALLED_APPS.remove('django.contrib.contenttypes')
            # INSTALLED_APPS.insert(0, 'bulb.contrib.contenttypes')
        else:
            pass
            # INSTALLED_APPS.insert(0, 'bulb.contrib.contenttypes')

    # Remove the native sessions module and set the bulb one.
    try:
        INSTALLED_APPS = root_settings['INSTALLED_APPS']

    except KeyError:
        root_settings['INSTALLED_APPS'] = ['bulb.contrib.sessions', ]

    else:
        if 'django.contrib.sessions' in INSTALLED_APPS:
            INSTALLED_APPS.remove('django.contrib.sessions')
            INSTALLED_APPS.insert(0, 'bulb.contrib.sessions')
        else:
            pass
            INSTALLED_APPS.insert(0, 'bulb.contrib.sessions')

    # Add the bulb database module.
    try:
        INSTALLED_APPS = root_settings['INSTALLED_APPS']

    except KeyError:
        root_settings['INSTALLED_APPS'] = ['bulb.db', ]

    else:
        INSTALLED_APPS.insert(0, 'bulb.db')

    # Add the bulb sftp and cdn module.
    try:
        INSTALLED_APPS = root_settings['INSTALLED_APPS']

    except KeyError:
        root_settings['INSTALLED_APPS'] = ['bulb.sftp_and_cdn', ]

    else:
        INSTALLED_APPS.insert(0, 'bulb.sftp_and_cdn')

    # Add the bulb contrib.statictools module.
    try:
        INSTALLED_APPS = root_settings['INSTALLED_APPS']

    except KeyError:
        root_settings['INSTALLED_APPS'] = ['bulb.contrib.statictools', ]

    else:
        INSTALLED_APPS.insert(0, 'bulb.contrib.statictools')

    # Add the bulb contrib.handling module.
    try:
        INSTALLED_APPS = root_settings['INSTALLED_APPS']

    except KeyError:
        root_settings['INSTALLED_APPS'] = ['bulb.contrib.handling', ]

    else:
        INSTALLED_APPS.insert(0, 'bulb.contrib.handling')

    # Add the bulb template module.
    try:
        INSTALLED_APPS = root_settings['INSTALLED_APPS']

    except KeyError:
        root_settings['INSTALLED_APPS'] = ['bulb.template', ]

    else:
        INSTALLED_APPS.insert(0, 'bulb.template')

    # Add the bulb utils module.
    try:
        INSTALLED_APPS = root_settings['INSTALLED_APPS']

    except KeyError:
        root_settings['INSTALLED_APPS'] = ['bulb.utils', ]

    else:
        INSTALLED_APPS.insert(0, 'bulb.utils')

    # Add the django-compressor module.
    try:
        INSTALLED_APPS = root_settings['INSTALLED_APPS']

    except KeyError:
        root_settings['INSTALLED_APPS'] = ['compressor', ]

    else:
        INSTALLED_APPS.insert(0, 'compressor')

    # Remove the native auth middleware and set the bulb one.
    try:
        MIDDLEWARE = root_settings['MIDDLEWARE']

    except KeyError:
        root_settings['MIDDLEWARE'] = ['django.contrib.auth.middleware.AuthenticationMiddleware', ]

    else:
        if 'django.contrib.auth.middleware.AuthenticationMiddleware' in MIDDLEWARE:
            MIDDLEWARE.remove('django.contrib.auth.middleware.AuthenticationMiddleware')
            MIDDLEWARE.insert(2, 'bulb.contrib.auth.middleware.AuthenticationMiddleware')
        else:
            MIDDLEWARE.insert(2, 'bulb.contrib.auth.middleware.AuthenticationMiddleware')

    # Remove the native sessions middleware and set the bulb one.
    try:
        MIDDLEWARE = root_settings['MIDDLEWARE']

    except KeyError:
        root_settings['MIDDLEWARE'] = ['bulb.contrib.sessions.middleware.SessionMiddleware', ]

    else:
        if 'django.contrib.sessions.middleware.SessionMiddleware' in MIDDLEWARE:
            MIDDLEWARE.remove('django.contrib.sessions.middleware.SessionMiddleware')
            MIDDLEWARE.insert(1, 'bulb.contrib.sessions.middleware.SessionMiddleware')
        else:
            MIDDLEWARE.insert(1, 'bulb.contrib.sessions.middleware.SessionMiddleware')

    # Set the bulb's HandlingMiddleware.
    try:
        MIDDLEWARE = root_settings['MIDDLEWARE']

    except KeyError:
        root_settings['MIDDLEWARE'] = ['bulb.contrib.handling.middleware.HandlingMiddleware']

    else:
        MIDDLEWARE.insert(1, 'bulb.contrib.handling.middleware.HandlingMiddleware')

    # Remove the native auth context processcors.
    try:
        TEMPLATES_context_processors = root_settings['TEMPLATES'][0]['OPTIONS']['context_processors']

    except KeyError:
        root_settings['TEMPLATES'][0]['OPTIONS']['context_processors'] = ['bulb.contrib.auth.context_processors.user_context_processors',]

    else:
        if 'django.contrib.auth.context_processors.auth' in TEMPLATES_context_processors:
            TEMPLATES_context_processors.remove('django.contrib.auth.context_processors.auth')
            TEMPLATES_context_processors.insert(0, 'bulb.contrib.auth.context_processors.user_context_processors')
        else:
            pass
            TEMPLATES_context_processors.insert(0, 'bulb.contrib.auth.context_processors.user_context_processors')

    # Add the bulb template 'debug' context processors.
    try:
        TEMPLATES_context_processors = root_settings['TEMPLATES'][0]['OPTIONS']['context_processors']

    except KeyError:
        root_settings['TEMPLATES'][0]['OPTIONS']['context_processors'] = ['bulb.template.context_processors.debug',]

    else:
        TEMPLATES_context_processors.insert(0, 'bulb.template.context_processors.debug')

    # Add the bulb admin 'additionnal_admin_modules' context processors.
    try:
        TEMPLATES_context_processors = root_settings['TEMPLATES'][0]['OPTIONS']['context_processors']

    except KeyError:
        root_settings['TEMPLATES'][0]['OPTIONS']['context_processors'] = [
            'bulb.contrib.admin.context_processors.additionnal_admin_modules', ]

    else:
        TEMPLATES_context_processors.insert(0, 'bulb.contrib.admin.context_processors.additionnal_admin_modules')

    # Remove the DATABASES variable :
    try:
        DATABASES = root_settings['DATABASES']

    except KeyError:
        pass

    else:
        del root_settings["DATABASES"]


    ##################
    # AUTHENTICATION #
    ##################

    # Remove the native auth backends config.
    root_settings['AUTH_USER_MODEL'] = None
    root_settings['AUTHENTICATION_BACKENDS'] = []
    root_settings['AUTH_PASSWORD_VALIDATORS'] = []

    ############################################################################
    #                        bulb's variables                      #
    ############################################################################

    ################
    # BULB DATABASE #
    ################

    # This parameter defines the attempts number where the application will retry to establish the initial connection
    # with the Neo4j database (each attempt is separated by 1 second).
    # It prevents failed when the database takes a long time to start or restart.
    # Explanations here :
    # https://neo4j.com/docs/operations-manual/3.0/deployment/post-installation/#post-installation-wait-for-start

    if root_settings['DEBUG']:
        root_settings['BULB_INITIAL_CONNECTION_ATTEMPTS_NUMBER'] = 2
    else:
        root_settings['BULB_INITIAL_CONNECTION_ATTEMPTS_NUMBER'] = 8

    # Neo4j authentication :
    root_settings['BULB_DATABASE_URI'] = "bolt://localhost:7687"

    root_settings['BULB_DATABASE_ID'] = 'neo4j'

    root_settings['BULB_DATABASE_PASSWORD'] = 'neo4j'

    # Encrypting trafic between the Neo4j driver and the Neo4j instance.
    # Explanations here : https://neo4j.com/docs/developer-manual/3.0/drivers/driver/#driver-authentication-encryption
    root_settings['BULB_DATABASE_ENCRYPTED'] = True

    # Verification against "man-in-the-middle" attack.
    # Explanations : https://neo4j.com/docs/developer-manual/3.0/drivers/driver/#_trust
    # Choices :
    # 0 : TRUST_ON_FIRST_USE     (Deprecated)
    # 1 : TRUST_SIGNED_CERTIFICATES     (Deprecated)
    # 2 : TRUST_ALL_CERTIFICATES
    # 3 : TRUST_CUSTOM_CA_SIGNED_CERTIFICATES
    # 4 : TRUST_SYSTEM_CA_SIGNED_CERTIFICATES
    # 5 : TRUST_DEFAULT = TRUST_ALL_CERTIFICATES
    root_settings['BULB_DATABASE_TRUST'] = 2

    # These parameters define the transactions modalities (after the establishment of the initial connection)
    # Explanations here : https://neo4j.com/docs/api/python-driver/current/driver.html#max-connection-lifetime
    root_settings['BULB_DATABASE_MAX_CONNECTION_LIFETIME'] = 60 * 60

    root_settings['BULB_DATABASE_MAX_CONNECTION_POOL_SIZE'] = 50

    root_settings['BULB_DATABASE_CONNECTION_ACQUISITION'] = 60

    if root_settings["DEBUG"]:
        root_settings['BULB_DATABASE_CONNECTION_TIMEOUT'] = 0.1
    else:
        root_settings['BULB_DATABASE_CONNECTION_TIMEOUT'] = 15

    root_settings['BULB_DATABASE_MAX_RETRY_TIME'] = 15

    ################
    # BULB SESSIONS #
    ################

    root_settings['BULB_SESSION_CHANGE_ON_EVERY_REQUEST'] = False

    # You can define all the SESSION_ variables like with the native Django package.
    root_settings['SESSION_ENGINE'] = 'bulb.contrib.sessions.backends.db'

    root_settings['SESSION_SERIALIZER'] = 'bulb.contrib.sessions.serializers.JSONSerializer'

    root_settings['SESSION_COOKIE_AGE'] = 60 * 60 * 24

    ######################
    # BULB AUTHENTICATION #
    ######################

    root_settings['BULB_LOGIN_URL'] = '/auth/login/'

    root_settings['BULB_HOME_PAGE_URL'] = '/home'

    root_settings['BULB_LOGIN_REDIRECT_URL'] = '/home/'

    root_settings['BULB_LOGOUT_REDIRECT_URL'] = None  # TODO

    root_settings['BULB_USER_NODE_MODEL_FILE'] = "bulb.contrib.auth.node_models"

    root_settings['BULB_ANONYMOUSUSER_NODE_MODEL_FILE'] = "bulb.contrib.auth.node_models"

    root_settings['BULB_PERMISSION_NODE_MODEL_FILE'] = "bulb.contrib.auth.node_models"

    root_settings['BULB_GROUP_NODE_MODEL_FILE'] = "bulb.contrib.auth.node_models"

    ##############
    # BULB PEPPER #
    ##############

    root_settings['BULB_PEPPER_1'] = None

    root_settings['BULB_PEPPER_2'] = None

    ############
    # BULB SFTP #
    ############
    root_settings['BULB_USE_SFTP'] = False

    root_settings['BULB_SFTP_HOST'] = None

    root_settings['BULB_SFTP_USER'] = None

    root_settings['BULB_SFTP_PASSWORD'] = None

    root_settings['BULB_SFTP_HOST_SSH_KEY'] = None

    root_settings['BULB_SFTP_PULL_URL'] = None

    ############################
    # BULB SFTP SRC STATICFILES #
    ############################

    """
        Possible values :
            - "raw" : The 'staticfiles' folder will be pushed on the sftp.
            - "bundled" : The "bundled_staticfiles" folder will be created from the 'staticfiles' folder and will be pushed
                          on the sftp.
            - "both" : Both 'staticfiles' and 'bundled_staticfiles' folders will be pushed on the sftp.
    """
    root_settings['BULB_SFTP_SRC_STATICFILES_MODE'] = "both"

    """
        The webpack polyfill is the easiest way to implement polyfills for the entire website.
        Just set this variable on True and all your scripts will be compatible for all browsers of all versions.

        But to obtain a more powerful website, it is recommended to don't use the webpack polyfill because:
        - The webpack polyfill is directly implemented into your bundle scripts, this means that if one of your scripts doesn't 
          need all polyfills , they will still be loaded.
        - The webpack polyfill don't have any regard on the browser used to load a page. This means that if your page is open 
          on a modern browser which doesn't need any polyfills, they will still be loaded.

        Best solution : Use 'polyfill.io' which one will load a the polyfills required in each different context. 
        See : (TODO: Add the related documentation).

        A good configuration of polyfill.io : 
            <script crossorigin="anonymous" src="https://polyfill.io/v3/polyfill.min.js?flags=gated&features=blissfuljs%2Cdefault%2Ces2015%2Ces2016%2Ces2017%2Ces5%2Ces6%2Ces7"></script>
        (Implement this script tag as the first script tag of each page which could need some polyfills.)
    """
    root_settings['BULB_SRC_BUNDLES_USE_WEBPACK_POLYFILL'] = False

    ###########
    # BULB CDN #
    ###########

    # CDN77:
    root_settings['BULB_USE_CDN77'] = False

    root_settings['BULB_CDN77_LOGIN'] = None

    root_settings['BULB_CDN77_API_KEY'] = None

    root_settings['BULB_CDN77_RESOURCE_ID'] = None  # See : https://client.cdn77.com/support/api/version/2.0/cdn-resource#List

    #############
    # BULB ADMIN #
    #############

    root_settings['BULB_ADMIN_BASEPATH_NAME'] = "admin"

    # Format {"<application name>": {"printed_name": "xxx", "path_name": "xxx", "home_view_url_name": "xxx"}, ...}
    root_settings['BULB_ADDITIONAL_ADMIN_MODULES'] = {}

    ############################################################################
    #                      bulb used variables                     #
    ############################################################################

    # The first hasher in this list is the preferred algorithm.  Any password using different algorithms will be
    # converted automatically upon login.
    root_settings['PASSWORD_HASHERS'] = [
        'django.contrib.auth.hashers.BCryptSHA256PasswordHasher',
    ]

    if root_settings['BULB_SFTP_HOST'] is not None:
        root_settings['STATIC_URL'] = "https://" + root_settings['BULB_SFTP_HOST'] + "/"





from bulb.sftp_and_cdn.exceptions import BULBStaticfilesError
from bulb.utils import get_folders_paths_list
import bulb
from django.core.management.base import BaseCommand
from django.template import Template, Context
from django.conf import settings
import subprocess
import lxml.html
import uuid
import os

BASE_DIR = settings.BASE_DIR


class Command(BaseCommand):
    args = ''
    help = """
            Generate the bundled_staticfiles folder which contains all the bundle of the project staticfiles.
            """

    def handle(self, *args, **options):
        if not settings.DEBUG:
            """
            If the command is executed when DEBUG = False, errors can occur from the bulb templatetags.
            Indeed, if DEBUG = False, if {% static_raw_src %}, {% static_bundled_src %} are used, staticfiles will be
            recovered from the SFTP : this will cause that the 'collectstatic command will collect staticfiles from the 
            old files on the SFTP, and not from the local new files.
            But also in certain cases, bundled staticfiles will be used : this will cause that the new bundle files will 
            be build from the old bundle files.

            See : TODO : add the related documentation page url.
            """
            raise BULBStaticfilesError("The DEBUG variable must be set on 'True' to handle new staticfiles.")

        else:
            # Bundling temporary files paths :
            package_file_path = None
            package_lock_file_path = None
            babelrc_file_path = None
            entry_file_path = None
            webpack_config_file_path = None

            # Test if the staticfiles folder exist.
            if not os.path.isdir(os.path.join(BASE_DIR, "staticfiles")):
                raise BULBStaticfilesError("The 'staticfiles' folder was not found. Please run the 'python manage.py collectstatic' command.")

            else:
                try:
                    bulb_path = bulb.__path__[0]

                    # Create the package.json file.
                    package_file_path = os.path.join(BASE_DIR, "package.json")
                    package_lock_file_path = os.path.join(BASE_DIR, "package-lock.json")
                    package_file = open(package_file_path, "w")
                    package_file.write(open(bulb_path + "/sftp_and_cdn/webpack_files/package.json", "r").read())
                    package_file.close()

                    # Create the .babelrc file.
                    babelrc_file_path = os.path.join(BASE_DIR, ".babelrc")
                    babelrc_file = open(babelrc_file_path, "w")
                    babelrc_file.write(open(bulb_path + "/sftp_and_cdn/webpack_files/.babelrc", "r").read())
                    babelrc_file.close()

                    # Install npm dependencies
                    subprocess.call(f"npm install", shell=True)

                    # Get paths of all "pages" folders of the project.
                    pages_folders_paths = get_folders_paths_list("pages")

                    # Get the bundled_staticfiles folder path.
                    bundled_staticfiles_folder_path = os.path.join(BASE_DIR, "bundled_staticfiles")

                    # Remove the old bundled_staticfiles folder and create a new one.
                    subprocess.call(f"rm -rf {bundled_staticfiles_folder_path}", shell=True)
                    os.mkdir(bundled_staticfiles_folder_path)

                    print("\n--------------")
                    print("-- BUNDLING --")
                    print("--------------\n")

                    # Loop on each "pages" folders of the project.
                    for pages_folder_path in pages_folders_paths:

                        splitted_pages_folder_path = pages_folder_path.split("/")

                        # Get the parent folder name and build its future path in the bundled_staticfiles folder.
                        parent_folder_name = splitted_pages_folder_path[-2] if splitted_pages_folder_path[-2] != "templates" else splitted_pages_folder_path[-3]
                        parent_folder_path = bundled_staticfiles_folder_path + "/" + parent_folder_name

                        print("\n     " + ("-" * (len(parent_folder_name) + 6)))
                        print("     -- " + parent_folder_name.upper() + " --")
                        print("     " + ("-" * (len(parent_folder_name) + 6)) + "\n")

                        # Create the current parent folder in the bundled_staticifiles folder.
                        os.mkdir(parent_folder_path)

                        # Loop on each page of the current "pages" folder, get their dependencies and bundle them.
                        for root, dirs, files in os.walk(pages_folder_path):

                            for file in files:
                                file_path = pages_folder_path + "/" + file

                                print("\n           -", file)

                                template_content = None
                                with open(file_path, "r") as template_file:
                                    template_content = template_file.read()
                                    template_content = template_content.replace('{% url', 'xxx')

                                template_object = Template(template_content)
                                context = Context({"DEBUG": True})
                                template = template_object.render(context)
                                # template_object = get_template(file_path)
                                # template = template_object.render()
                                doc = lxml.html.document_fromstring(template)

                                # Create the entry.js file.
                                entry_file_path = os.path.join(BASE_DIR, "entry.js")
                                entry_file = open(entry_file_path, "a")

                                # Add polyfill dependencies to the entry.js file.
                                if settings.BULB_SRC_BUNDLES_USE_WEBPACK_POLYFILL:
                                    entry_file.write("import 'core-js';import 'regenerator-runtime/runtime';")

                                print("\n               Searching CSS dependencies...")
                                print("               Found :\n")

                                links = doc.xpath("//link[@rel='stylesheet']")

                                # Add CSS dependencies to the entry.js file.
                                for link in links:
                                    href_value = link.get("href")
                                    print("                 -", href_value)

                                    # Ignore href with external urls.
                                    if href_value[:4] == "http":
                                        continue

                                    else:
                                        related_staticfiles_path = "." + "/staticfiles" + href_value[7:]

                                        entry_file.write(f"import {'a' + uuid.uuid4().hex} from '{related_staticfiles_path}';")

                                # Add JS dependencies to the entry.js file.
                                print("\n               Searching JS dependencies...")
                                print("               Found :\n")

                                scripts = doc.xpath("//script")

                                for script in scripts:
                                    src_value = script.get("src")
                                    print("                 -", src_value)

                                    # Ignore href with external urls.
                                    if src_value is None:
                                        continue

                                    elif src_value[:4] == "http":
                                        continue


                                    else:
                                        related_staticfiles_path = "." + "/staticfiles" + src_value[7:]

                                        entry_file.write(f"import {'a' + uuid.uuid4().hex} from '{related_staticfiles_path}';")

                                entry_file.close()

                                print("\n               ------------------------")
                                print("               -- INITIALIZE WEBPACK --")
                                print("               ------------------------\n")

                                # Create webpack.config.js file.
                                webpack_config_file_path = os.path.join(BASE_DIR, "webpack.config.js")
                                webpack_config_file = open(webpack_config_file_path, "a")
                                webpack_config_file.write(f"""// Don't modify this file, it is generated by a Python script (see sftp_and_cdn.management.commands.handlestatic.py)
        process.env.WEBPACK_ENTRY = ['{entry_file_path}'];
        process.env.WEBPACK_OUTPUT = '{parent_folder_path}';
        process.env.BUNDLE_NAME = '{file[:-5]}';
        
        """)
                                webpack_config_file.write(open(bulb_path + "/sftp_and_cdn/webpack_files/webpack.config.js", "r").read())
                                webpack_config_file.close()

                                print("                     Done ✔ ")

                                print("\n               ---------------------")
                                print("               -- RUN WEBPACK --")
                                print("               ---------------------\n")

                                subprocess.call(f"npm run build", shell=True)

                                # Remove the entry.js file.
                                os.remove(entry_file_path)

                                # # Remove the webpack.config.js file.
                                os.remove(webpack_config_file_path)

                    # Remove the .babelrc file
                    os.remove(babelrc_file_path)

                    # # Remove the package.json and package-lock.json files.
                    os.remove(package_file_path)
                    os.remove(package_lock_file_path)

                    return bundled_staticfiles_folder_path

                # If any error occurs during bundling, all the temporary files are removed.
                except:

                    # Remove the entry.js file.
                    try:
                        os.remove(entry_file_path)

                    except (FileNotFoundError, TypeError):
                        pass

                    # # Remove the webpack.config.js file.
                    try:
                        os.remove(webpack_config_file_path)

                    except (FileNotFoundError, TypeError):
                        pass

                    # Remove the .babelrc file.
                    try:
                        os.remove(babelrc_file_path)

                    except (FileNotFoundError, TypeError):
                        pass

                    # # Remove the package.json and package-lock.json files.
                    try:
                        os.remove(package_file_path)

                    except FileNotFoundError:
                        pass

                    try:
                        os.remove(package_lock_file_path)

                    except (FileNotFoundError, TypeError):
                        pass

                    raise



from bulb.sftp_and_cdn.sftp import SFTP
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    args = ''
    help = """
            Clear (remove) the staticfiles folder on the SFTP server and purge the CDN if there is one.
            """

    def add_arguments(self, parser):
        parser.add_argument("-r", "--raw", action="store_true")
        parser.add_argument("-b", "--bundled", action="store_true")

    def handle(self, *args, **kwargs):
        if kwargs["raw"]:
            SFTP.clear_src_staticfiles(src_type="raw")

        if kwargs["bundled"]:
            SFTP.clear_src_staticfiles(src_type="bundled")

        if not kwargs["raw"] and not kwargs["bundled"]:
            SFTP.clear_src_staticfiles(src_type="raw")
            SFTP.clear_src_staticfiles(src_type="bundled")




from bulb.contrib.auth.decorators import login_required, protect_authentication_view, staff_only
from bulb.contrib.auth.authentication import preserve_or_login, authenticate

from bulb.contrib.auth.authentication import force_clear_user_session
from bulb.contrib.admin.forms import AdminLoginForm

from bulb.contrib.auth.node_models import AnonymousUser
from bulb.db import gdbh

from django.shortcuts import render, redirect
from django.http import JsonResponse
from django.conf import settings

login_page_url = "/" + settings.BULB_ADMIN_BASEPATH_NAME + "/login"
home_page_url = "/" + settings.BULB_ADMIN_BASEPATH_NAME + "/"


@staff_only()
@login_required(login_page_url=login_page_url)
def admin_home_view(request):
    return render(request, "admin/pages/admin_home.html")


@staff_only()
@protect_authentication_view(home_page_url=home_page_url)
def admin_login_view(request):
    form = AdminLoginForm(request.POST or None)

    if form.is_valid():
        email = form.cleaned_data["email"]
        password = form.cleaned_data["password"]

        user = authenticate(email, password)

        if AnonymousUser in user.__mro__:
            preserve_or_login(request, user)

            if "next" in request.POST:
                return redirect(f"http://127.0.0.1:8000{request.POST.get('next')}", "/admin")

            else:
                return redirect("/admin")

    if request.is_ajax():
        response = gdbh.r_transaction("""
                   MATCH (u:User {%s: '%s'})
                   RETURN (u)
                   """ % (request.POST["field_id"], request.POST["field_value"]))

        if response:
            return JsonResponse({"value_is_unused": False})
        else:
            return JsonResponse({"value_is_unused": True})

    return render(request, "admin/pages/admin_login.html", {"form": form})


@staff_only()
def admin_logout_view(request):
    if "next" in request.POST:
        force_clear_user_session(request)
        return redirect(f"{login_page_url}/?next={request.POST.get('next')}")

    else:
        force_clear_user_session(request)
        return redirect(login_page_url)



from bulb.db.exceptions import BULBNodeError

class BULBSessionError(Exception):
    pass


class BULBSessionDoesNotExist(BULBNodeError):
    pass


class BULBSessionDoesNotHaveData(BULBNodeError):
    pass



from bulb.db.exceptions import BULBNodeLabelsInitializationError
from bulb.db.node_models import Node
from bulb.utils import get_files_paths_list
from bulb.db.base import gdbh
from django.core.management.base import BaseCommand
import importlib.util
import os

BASE_DIR = os.environ["BASE_DIR"]


class Command(BaseCommand):
    args = ''
    help = """
            Apply constraints of node models properties.
            """

    def handle(self, *args, **options):
        nodes_models_files_paths = get_files_paths_list("node_models.py")

        # beginning CONSOLE RENDER PART 1 #
        print("\n--------------------------------------\n")

        found_path_number = len(nodes_models_files_paths)

        if found_path_number == 1:
            print(f"    {len(nodes_models_files_paths)} 'node_models.py' file has been found :")

        if found_path_number > 1:
            print(f"    {len(nodes_models_files_paths)} 'node_models.py' files have been found :")

        else:
            print(f"    No one file named 'node_models.py' has been found in this django project.")
        # end CONSOLE RENDER PART 1 #

        # For each path of the node_models.py files :
        for path in nodes_models_files_paths:

            # beginning CONSOLE RENDER PART 2 #
            print(f"\n        -> '{path}' :")
            # end CONSOLE RENDER PART 2 #

            # Import the module from his path
            spec = importlib.util.spec_from_file_location("node_models", path)
            node_models = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(node_models)

            # Get all the module classes that have in their parents the Node class.
            node_classes = []
            for k, v in node_models.__dict__.items():
                try:
                    if Node in v.__mro__:
                        node_classes.append(v)
                except:
                    pass

            # beginning CONSOLE RENDER PART 3 #
            if len(node_classes) == 1:
                print(f"                1 node class named has been found : ")

            elif len(node_classes) > 1:
                print(f"                {len(node_classes)} node classes have been found : ")

            else:
                print(f"                No one node class (inherting from Node) has been found.")
            # end CONSOLE RENDER PART 3 #

            # For each node class in the current module node classes :
            for node_class in node_classes:
                node_class_name = node_class.__name__

                # Explanation : __module__ return only the name of the module where is contained the node_class (here :
                # "node_models"), but if the node_class is an import, __module__ return the full module path (example :
                # "bulb.contrib.sessions.node_models"). So, this line prevent the detection of the imported
                # classes in the node_models files.
                if node_class.__module__ == "node_models":
                    # beginning CONSOLE RENDER PART 4 #
                    print(f"                    - {node_class_name}")
                    # end CONSOLE RENDER PART 4 #

                    # Get needed datas (labels and properties) and format the labels dict into cypher format
                    node_class_labels_list = node_class._get_labels()
                    node_class_labels_cypher_format = self.format_labels_to_cypher(node_class_labels_list, node_class_name)
                    node_class_properties = node_class._get_property_fields()

                    # For each property of the current node class
                    for property_name, property_content in node_class_properties.items():

                        # If the property is defined as "required" (required=True), create REQUIRED constraint in the database
                        if property_content.required:
                            gdbh.w_transaction("""
                                CREATE CONSTRAINT ON (x:%s) 
                                ASSERT exists(x.%s)
                                """ % (node_class_labels_cypher_format, property_name))

                            # beginning CONSOLE RENDER PART 5 #
                            print(f"                        ✔   Apply REQUIRED constraint on '{property_content.key}'.")
                            # end CONSOLE RENDER PART 5 #

                        # If the property is defined as "unique" (unique=True), create UNIQUE constraint in the database
                        if property_content.unique:
                            gdbh.w_transaction("""
                                CREATE CONSTRAINT ON (x:%s)
                                ASSERT x.%s IS UNIQUE
                                """ % (node_class_labels_cypher_format, property_content.key))

                            # beginning CONSOLE RENDER PART 6 #
                            print(f"                        ✔   Apply UNIQUE constraint on '{property_content.key}'.")
                            # end CONSOLE RENDER PART 6 #

        # beginning CONSOLE RENDER PART 7 #
        print("\n--------------------------------------\n")
        # end CONSOLE RENDER PART 7 #

    @staticmethod
    # Convert dict format to a cypher labels format
    def format_labels_to_cypher(labels_list, label_class_name):
        render = []

        if not labels_list or labels_list is None:
            return label_class_name

        else:
            if isinstance(labels_list, list):
                for label in labels_list:
                    render.append(label)
                return ':'.join(render)

            else:
                raise BULBNodeLabelsInitializationError('self.labels attribute must be a list.')


from bulb.contrib.handling.views import handling_home_view, node_model_home_view, node_handling_view, node_creation_view
from django.urls import path, re_path

urlpatterns = [
    path("", handling_home_view, name="handling_home"),
    re_path(r"^(?P<node_model_name>[a-zA-Z_]+)$", node_model_home_view, name="node_model_home"),
    re_path(r"^(?P<node_model_name>[a-zA-Z_]+)/(?P<node_uuid>([a-z0-9]{32}|None))$", node_handling_view, name="node_handling"),
    re_path(r"^(?P<node_model_name>[a-zA-Z_]+)/creation", node_creation_view, name="node_creation")

]


from urllib.parse import quote, urljoin
from django import template
from django.conf import settings
import os

register = template.Library()

BASE_DIR = os.environ["BASE_DIR"]


class BULBStaticTemplateTagsError(Exception):
    pass


@register.simple_tag()
def static_raw_src(path):
# def static_src(path):
    if settings.DEBUG:
        return urljoin(settings.STATIC_URL, quote(path))

    else:
        pull_url = settings.BULB_SFTP_PULL_URL

        if pull_url is not None:
            src_url = pull_url + "/staticfiles/raw_src/"
            return urljoin(src_url, quote(path))

        else:
            raise BULBStaticTemplateTagsError(
                "To use the '{% static_src %}' tag you must define the BULB_SFTP_PULL_URL variable in your 'settings.py' file.")


@register.simple_tag()
def static_bundled_src(path):
    if settings.DEBUG:
        return urljoin(os.path.join(BASE_DIR, "bundled_staticfiles"), quote(path))

    else:
        pull_url = settings.BULB_SFTP_PULL_URL

        if pull_url is not None:
            src_url = pull_url + "/staticfiles/bundled_src/"
            return urljoin(src_url, quote(path))

        else:
            raise BULBStaticTemplateTagsError(
                "To use the '{% static_src %}' tag you must define the BULB_SFTP_PULL_URL variable in your 'settings.py' file.")

@register.simple_tag()
def static_content(path):
    if settings.DEBUG or not settings.BULB_USE_SFTP:
        return urljoin(settings.STATIC_URL, quote(path))

    else:
        pull_url = settings.BULB_SFTP_PULL_URL

        if pull_url is not None:
            src_url = pull_url + "/staticfiles/content/"
            return urljoin(src_url, quote(path))

        else:
            raise BULBStaticTemplateTagsError(
                "To use the '{% static_content %}' tag you must define the BULB_SFTP_PULL_URL variable in your 'settings.py' file.")



from bulb.contrib.sessions.exceptions import BULBSessionError, BULBSessionDoesNotExist, BULBSessionDoesNotHaveData
from bulb.db import node_models
from bulb.db.base import gdbh
from django.utils import timezone


class RelatedUserRelationship(node_models.Relationship):
    rel_type = "IS_SESSION_OF"
    direction = "from"
    start = "self"
    target = "User"
    auto = False
    on_delete = "PROTECT"
    unique = True


class AbstractBaseSession:

    def encode(self, session_dict):
        """
        Return the given session dictionary serialized and encoded as a string.
        """
        session_store_class = self.__class__.get_session_store_class()
        return session_store_class().encode(session_dict)

    def save(self, session_key, session_dict, expire_date):
        if session_dict:
            s = self.__class__(session_key, self.encode(session_dict), expire_date)
        else:
            raise BULBSessionDoesNotHaveData("The does'nt have datas ('session_dict'), failed to save it.")
        return s

    def __str__(self):
        return self.session_key

    @classmethod
    def get_session_store_class(cls):
        raise NotImplementedError

    def get_decoded(self):
        session_store_class = self.get_session_store_class()
        return session_store_class().decode(self.session_data)


class Session(node_models.Node, AbstractBaseSession):
    """
    Django provides full support for anonymous sessions. The session
    framework lets you store and retrieve arbitrary data on a
    per-site-visitor basis. It stores data on the server side and
    abstracts the sending and receiving of cookies. Cookies contain a
    session ID -- not the data itself.

    The Django sessions framework is entirely cookie-based. It does
    not fall back to putting session IDs in URLs. This is an intentional
    design decision. Not only does that behavior make URLs ugly, it makes
    your site vulnerable to session-ID theft via the "Referer" header.

    For complete documentation on using Sessions in your code, consult
    the sessions documentation that is shipped with Django (also available
    on the Django Web site).
    """

    session_key = node_models.Property(required=True,
                                       unique=True)  # TODO : add max_length = 40

    session_data = node_models.Property()

    expire_date = node_models.Property()  # TODO : add db_index = True

    related_user = RelatedUserRelationship()

    def __str__(self):
        return f'<Session object(session_key="{self.session_key}", expire_date="{self.expire_date}")>'

    def __repr__(self):
        return f'<Session object(session_key="{self.session_key}", expire_date="{self.expire_date}")>'

    @classmethod
    def get_session_store_class(cls):
        from bulb.contrib.sessions.backends.db import SessionStore
        return SessionStore

    @classmethod
    def get(cls, uuid=None, session_key=None, order_by=None, limit=None, skip=None, desc=False, only=None):
        """
        This method allow the retrieving of Session (or of one of its children classes) instances.


        :param uuid: The Universal Unique Identifier of a session to get an unique session instance.

        :param session_key: The session_key of a session to get an unique session instance.

        :param order_by: Must be the name of the property with which the returned datas will be sorted.
                         Examples : "datetime", "first_name", etc...

        :param limit: Must be an integer. This parameter defines the number of returned elements.

        :param skip: Must be an integer. This parameter defines the number of skipped elements. For example if
                     self.skip = 3, the 3 first returned elements will be skipped.

        :param desc: Must be a boolean. If it is False the elements will be returned in an increasing order, but it is
                     True, they will be returned in a descending order.

        :param only: Must be a list of field_names. If this parameter is filled, the return will not be Permission
                     instances, but a dict with "only" the mentioned fields.

        :return: If uuid is None, a list will be returned. Else it will be a unique instance.
        """
        property_statement = ""
        order_by_statement = ""
        limit_statement = ""
        skip_statement = ""
        desc_statement = ""

        # Build the property_statement.
        if uuid is not None and session_key is not None:
            property_statement = "{uuid:'%s', session_key:'%s'}" % (uuid, session_key)

        elif uuid is not None:
            property_statement = "{uuid:'%s'}" % uuid

        elif session_key is not None:
            property_statement = "{session_key:'%s'}" % session_key

        # Build the match_statement.
        cyher_labels = node_models.DatabaseNode.format_labels_to_cypher(cls._get_labels())
        match_statement = f"MATCH (s:{cyher_labels} {property_statement})"

        # Build the with_statement.
        with_statement = "WITH s"

        # Build order_by statements.
        if order_by is not None:
            order_by_statement = f"ORDER BY s.{order_by}"

        # Build return_statement statements.
        if not only:
            return_statement = "RETURN (s)"

        else:
            only_statement_list = []

            for element in only:
                only_statement_list.append(f"s.{element}")

            only_statement = ", ".join(only_statement_list)

            return_statement = f"RETURN {only_statement}"

        # Build limit_statement.
        if limit is not None:
            if not isinstance(limit, str) and not isinstance(limit, int):
                raise BULBSessionError(
                    f"The 'limit' parameter of the get() method of {cls.__name__} must be a string or an integer.")

            else:
                limit_statement = f"LIMIT {limit}"

        # Build skip_statement and add its required variable.
        if skip is not None:
            if not isinstance(skip, str) and not isinstance(skip, int):
                raise BULBSessionError(
                    f"The 'skip' parameter of the get() method of {cls.__name__} must be a string or an integer.")

            else:
                skip_statement = f"SKIP {skip}"

        # Build desc_statement.
        if not isinstance(desc, bool):
            raise BULBSessionError(
                f"The 'desc' parameter of the get() method of {cls.__name__} must be a boolean.")

        else:
            if desc is True:
                desc_statement = "DESC"

        request_statement = """
             %s
             %s
             %s
             %s
             %s
             %s
             %s
             """ % (match_statement,
                    with_statement,
                    order_by_statement,
                    desc_statement,
                    limit_statement,
                    skip_statement,
                    return_statement)

        response = gdbh.r_transaction(request_statement)

        if response:
            if only is None:
                fake_instances_list = []

                for session_object in response:
                    fake_instances_list.append(cls.build_fake_instance(session_object["s"],
                                                                       forced_fake_instance_class=cls))

                if uuid is not None or session_key is not None:
                    return fake_instances_list[0]

                else:
                    return fake_instances_list

            else:
                return response

        else:
            return None

    @classmethod
    def exists(cls, session_key):
        response = cls.get(session_key=session_key)

        if response:
            return True

        else:
            return False

    @classmethod
    def delete_session(cls, session_key):
        response = cls.get(session_key=session_key)

        if response:
            gdbh.w_transaction("MATCH (s:Session {session_key:'%s'}) DETACH DELETE (s)" % session_key)

        else:
            raise BULBSessionDoesNotExist(f"No session with session_key = '{session_key}'. So it cannot be deleted.")

    @classmethod
    def clear_expired_sessions(cls):
        gdbh.r_transaction("""
            MATCH (s:Session) 
            WHERE s.expire_date < datetime('%s') 
            DETACH DELETE (s)
            """ % str(timezone.now()).replace(' ', 'T'))




# Base bulb error
class bulbError(BaseException):
    pass


#  Base bulb error --> Client error
class BULBClientError(bulbError):
    pass


#  Base bulb error --> Client error --> Node error
class BULBNodeError(BULBClientError):
    pass

class BULBNodeWarning(Warning):
    pass

#  Base bulb error --> Client error --> Node error --> Node initialization error --> labels
class BULBNodeLabelsInitializationError(BULBNodeError):
    pass


# Base bulb error --> Client error --> Connection error
class BULBConnectionError(BULBClientError):
    pass


# Base bulb error --> Client error --> Connection warning
class BULBConnectionWarning(BULBClientError, Warning):
    pass


# Base bulb error --> Client error --> Transaction error
class BULBTransactionError(BULBClientError):
    pass


# Base bulb error --> Client error --> Transaction error --> Transaction type error
class BULBTransactionTypeError(BULBTransactionError):
    pass


# Base bulb error --> Client error --> Transaction error --> Transaction conflict error
class BULBTransactionConflictError(BULBTransactionError):
    pass


# Base bulb error --> Client error --> Session error
class BULBSessionError(BULBClientError):
    pass


#  Base bulb error --> Database error
class BULBDatabaseError(bulbError):
    pass


#  Base bulb error --> Database error --> field error
class BULBFieldError(BULBDatabaseError):
    pass


#  Base bulb error --> Database error --> field error --> required constraint error
class BULBRequiredConstraintError(BULBFieldError):
    pass


#  Base bulb error --> Database error --> unique constraint error
class BULBUniqueConstraintError(BULBFieldError):
    pass


#  Base bulb error --> Database error --> attribute error
class BULBAttributeError(BULBFieldError):
    pass


class BULBFileError(Exception):
    pass


class BULBRelationshipError(Exception):  # verified
    pass


class BULBRelationshipInstanceWarning(Warning):  # verified
    pass


class BULBBaseNodeAndRelationshipError(Exception):  # verified
    pass



from bulb.contrib.auth.exceptions import BULBLoginError
from bulb.contrib.auth.node_models import get_user_node_model, get_anonymoususer_node_model
from bulb.contrib.auth.hashers import _check_password
from bulb.db.base import gdbh
from django.conf import settings
import os

BASE_DIR = os.environ["BASE_DIR"]

User = get_user_node_model()
AnonymousUser = get_anonymoususer_node_model()


def authenticate(email, password):
    user = User.get(email=email)

    if not isinstance(user, AnonymousUser):
        if _check_password(password, user.password):
            if user.is_active_user:
                return user

    return AnonymousUser()


def _login_user(request, user):
    request.session['logged_user_uuid'] = user.uuid
    request.session.save()

    request.user = user

    gdbh.w_transaction("""
        MATCH (u:User {uuid: '%s'}), (s:Session {session_key: '%s'})
        CREATE (s)-[:IS_SESSION_OF]->(u)
        """ % (user.uuid, request.session.session_key))


def preserve_or_login(request, if_authentication_user=AnonymousUser()):
    from_authentication_user = AnonymousUser()
    from_request_user = AnonymousUser()

    try:
        if not isinstance(if_authentication_user, AnonymousUser):
            if isinstance(if_authentication_user, User):
                from_authentication_user = if_authentication_user

        if not isinstance(request.user, AnonymousUser):
            if isinstance(request.user, User):
                from_request_user = request.user

        if not if_authentication_user and not request.user:
            raise BULBLoginError("To login an user with the 'preserve_or_login()' function, you must provide as function's parameters : the request and the user object.")

    except BULBLoginError:
        request.session.flush()
        raise

    else:
        # If the function is called during an authentication, to login an user :
        if not isinstance(from_authentication_user, AnonymousUser):

            # If the user is not an AnonymousUser and if a 'logged_user_uuid' is already in the session (if the
            # user try to login although it is MAYBE already) :
            if 'logged_user_uuid' in request.session:

                if not isinstance(from_request_user, AnonymousUser) and request.session['logged_user_uuid'] is not None:

                    # If the user is the same in database and session :
                    if from_request_user.uuid == from_authentication_user.uuid == request.session['logged_user_uuid']:
                        current_user_session = from_authentication_user.session.get()[0]

                        # If the session is the same in database and in cookie :
                        if current_user_session and request.session.session_key \
                                and current_user_session.session_key == request.session.session_key:

                            # If settings.BULB_SESSION_CHANGE_ON_EVERY_REQUEST is True, delete and create a new session
                            # Else, just preserve the current session.
                            if settings.BULB_SESSION_CHANGE_ON_EVERY_REQUEST:
                                force_clear_user_session(request)
                                _login_user(request, from_authentication_user)

                        # If the sessions datas are not identical
                        else:
                            force_clear_user_session(request)
                            _login_user(request, from_authentication_user)

                    # If the users datas are not identical
                    else:
                        force_clear_user_session(request)
                        _login_user(request, from_authentication_user)

                # Else if the user is not already logged in :
                else:
                    force_clear_user_session(request)
                    _login_user(request, from_authentication_user)

            else:
                force_clear_user_session(request)
                _login_user(request, from_authentication_user)

        # If the user is not provided during authentication (he is the currently logged user stored in request.user)
        elif not isinstance(from_request_user, AnonymousUser):

            # If a 'logged_user_uuid' is already in the session (if the user is MAYBE already logged) :
            if 'logged_user_uuid' in request.session:

                if request.session['logged_user_uuid'] is not None:

                    # If the user is the same in database and session :
                    if from_request_user.uuid == request.session['logged_user_uuid']:
                        session_request = from_request_user.session.get()
                        current_user_session = None

                        if session_request is not None:
                            current_user_session = session_request[0]

                        # If the session is the same in database and in cookie :
                        if current_user_session and request.session.session_key \
                                and current_user_session.session_key == request.session.session_key:

                            # If settings.BULB_SESSION_CHANGE_ON_EVERY_REQUEST is True, delete and create a new session
                            # Else, just preserve the current session.
                            if settings.BULB_SESSION_CHANGE_ON_EVERY_REQUEST:
                                force_clear_user_session(request)
                                _login_user(request, from_request_user)

                        # If the sessions datas are not identical
                        else:
                            force_clear_user_session(request)

                    # If the users datas are not identical
                    else:
                        force_clear_user_session(request)

                # Else if the user is not already logged in :
                else:
                    force_clear_user_session(request)

            else:
                force_clear_user_session(request)

        else:
            request.session.flush()


# Logout function : clear the sessions stored in the cookie and in the database + set request.user with an instance
# of AnonymousUsers
def force_clear_user_session(request):
    gdbh.w_transaction("""
        MATCH (:User {uuid: '%s'})<-[:IS_SESSION_OF]-(s:Session)
        DETACH DELETE (s)
        """ % (request.user.uuid))

    request.session.flush()
    request.user = AnonymousUser()
    request.session.setdefault('logged_user_uuid', None)


def get_user_model():
    return User


# Return the user from the request, if there is'nt, return an instance of AnonymousUser
def get_user_from_request(request):
    try:
        if 'logged_user_uuid' in request.session:

            if request.session['logged_user_uuid'] is not None:
                logged_user_uuid = request.session['logged_user_uuid']

                if logged_user_uuid is not None:

                    logged_user = User.get(uuid=logged_user_uuid)

                    if not isinstance(logged_user, AnonymousUser):
                        return logged_user

                    else:
                        # If the user does not exist, clear the session datas, that has nothing to do here
                        request.session.flush()
                        return AnonymousUser()

                else:
                    request.session.flush()
                    return AnonymousUser()

            else:
                request.session.flush()
                return AnonymousUser()

        else:
            request.session.flush()
            return AnonymousUser()

    except (AttributeError, TypeError):
        request.session.flush()
        return AnonymousUser()


# Return True if an user is logged, else return False
def get_user_is_logged(request):
    # if not NativeAnonymousUser in request.user.__class__.__mro__:
    if not AnonymousUser in request.user.__class__.__mro__:
        session_request = request.user.session.get()

        if session_request is not None:
            current_user_session = session_request[0]

            if current_user_session is not None:

                if request.session:

                    if "logged_user_uuid" in request.session:

                        logged_user_uuid = request.session['logged_user_uuid']
                        if logged_user_uuid is not None:

                            if request.user.uuid == logged_user_uuid and \
                                    request.session.session_key == current_user_session.session_key:

                                return True
    return False




import uuid


def make_uuid():
    return uuid.uuid4().hex


def compare_different_modules_classes(class1, class2):
    class2.__module__ = class1.__module__
    if class1.__name__ == class2.__name__ and class1.__dict__.keys() == class2.__dict__.keys():
        return True
    return False



from bulb.sftp_and_cdn.cdn_apis import CDN77
from bulb.sftp_and_cdn.exceptions import *
from django.conf import settings
from base64 import decodebytes
import warnings
import paramiko
import pysftp
import os

BASE_DIR = os.environ["BASE_DIR"]


class SFTP:

    @staticmethod
    def connect(log=None):

        keydata = settings.BULB_SFTP_HOST_SSH_KEY

        if keydata is None:
            raise BULBHostSSHKeyError("To establish an SFTP connection you have to provide the SSH key of the server. Please put it in the BULB_SFTP_HOST_SSH_KEY variable in 'settings.py'.")

        else:
            key = paramiko.RSAKey(data=decodebytes(keydata))

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=UserWarning)
                cnopts = pysftp.CnOpts()
                cnopts.hostkeys.add(settings.BULB_SFTP_HOST, 'ssh-rsa', key)

            return pysftp.Connection(host=settings.BULB_SFTP_HOST, username=settings.BULB_SFTP_USER, password=settings.BULB_SFTP_PASSWORD, cnopts=cnopts, log=log)

    @staticmethod
    def push_src_staticfiles(src_type, local_staticfiles_folder_path):
        """
        This method post new source staticfiles on the SFTP server.

        :param (required) src_type: The type of source staticfiles to post on the sftp. "src" / "bundled"
        :param (required) local_staticfiles_folder_path: The local staticfiles folder path.
        :return:
        """
        if src_type == "raw":
            sftp_staticfiles_folder_path = "/www/staticfiles/raw_src"

        elif src_type == "bundled":
            sftp_staticfiles_folder_path = "/www/staticfiles/bundled_src"

        else:
            raise BULBStaticfilesError(
                f'The "src_type" parameter of the clear_staticfiles() method should be "raw" or "bundled". It is {str(src_type)}.')

        with SFTP.connect() as sftp:

            print("\n------------------------------------------" + ("---" if src_type == "raw" else "-------"))
            print(f"--  UPLOAD NEW {src_type.upper()} SRC STATICFILES ON SFTP --")
            print(("---" if src_type == "raw" else "-------") + "------------------------------------------\n")

            print("\nThis operation could take few minutes (it depends of the number and the weight of your static files)...\n")

            try:
                sftp.put_r(local_staticfiles_folder_path, sftp_staticfiles_folder_path)

            # Check and create default storage folders if they are not already created.
            except FileNotFoundError:
                if not sftp.exists("/www/staticfiles"):
                    sftp.mkdir("/www/staticfiles")

                if not sftp.exists(f"/www/staticfiles/{src_type}_src"):
                    sftp.mkdir(f"/www/staticfiles/{src_type}_src")

                sftp.put_r(local_staticfiles_folder_path, sftp_staticfiles_folder_path)

            finally:
                print("Done ✔\n")


    @staticmethod
    def clear_src_staticfiles(src_type):
        """
        This method delete old source staticfiles on the SFTP server.

        :param (required) src_type: The type(s) of source staticfiles to remove. "src" / "bundled"
        :return:
        """
        if src_type == "raw":
            sftp_staticfiles_folder_path = "/www/staticfiles/raw_src"

        elif src_type == "bundled":
            sftp_staticfiles_folder_path = "/www/staticfiles/bundled_src"

        else:
            raise BULBStaticfilesError(f'The "src_type" parameter of the clear_staticfiles() method should be "raw", "bundled". It is {str(src_type)}.')

        with SFTP.connect() as sftp:

            files_to_purge_list = []

            def remove_file(path):
                if sftp.isfile(path):
                    rectified_path = "".join([x for x in path][4:])  # Delete '/www' on the path
                    files_to_purge_list.append(rectified_path)

                    sftp.remove(path)
                    print("'" + path + "'", " removed.")

            paths_list = []

            def add_to_paths_list(x):
                paths_list.append(x)

            print(f"\n-----------------------------------------{'----' if src_type == 'bundled' else ''}")
            print(f"--  REMOVE OLD {src_type.upper()} SRC FILES FROM SFTP --")
            print(f"-----------------------------------------{'----' if src_type == 'bundled' else ''}\n")

            try:
                sftp.walktree(sftp_staticfiles_folder_path, fcallback=remove_file, dcallback=remove_file, ucallback=remove_file)

            except FileNotFoundError:
                print(f"\n      There were no {src_type} source files to remove on the sftp.\n")

            print(f"\n-----------------------------------------------{'----' if src_type == 'bundled' else ''}")
            print(f"--  REMOVE OLD {src_type.upper()} SRC DIRECTORIES FROM SFTP --")
            print(f"-----------------------------------------------{'----' if src_type == 'bundled' else ''}\n")

            try:
                sftp.walktree(sftp_staticfiles_folder_path, fcallback=add_to_paths_list, dcallback=add_to_paths_list, ucallback=add_to_paths_list)

            except FileNotFoundError:
                print(f"\n      There were no {src_type} source directories to remove on the sftp.\n")

            else:
                while paths_list:
                    for path in paths_list:
                        try:
                            sftp.rmdir(path)
                            pass

                        except :
                            continue

                        else:
                            print(path, " removed !")
                            paths_list.remove(path)

            if settings.BULB_USE_CDN77:
                print(f"\n----------------------------------------------{'----' if src_type == 'bundled' else ''}")
                print(f"--  PURGE {src_type.upper()} STATICFILES ON CDN77 SERVERS  --")
                print(f"----------------------------------------------{'----' if src_type == 'bundled' else ''}\n")

                files_to_purge_number = len(files_to_purge_list)

                print(f"\n      {files_to_purge_number} file{'s' if files_to_purge_number > 1 or files_to_purge_number == 0 else ''} to purge.\n")

                # CDN77 allows up to 20 requests with up to 2000 files to purge every 5 minutes.
                # See : https://client.cdn77.com/support/api/version/2.0/data

                if files_to_purge_number != 0:

                    # If the number of files to purge is under the limit send a unique request:
                    if files_to_purge_number < 2000:
                        CDN77.purge(files_to_purge_list)

                    # Else send multiple requests, each with up to 2000 files maximum.
                    else:
                        request_number = 1

                        while files_to_purge_number >= 2000:
                            purge_response = CDN77.purge(files_to_purge_list[:2000], request_number)

                            request_number += 1
                            files_to_purge_number -= 2000
                            files_to_purge_list = files_to_purge_list[2000:]

                            if purge_response.status_code != "200":
                                break

                        if files_to_purge_number > 0:
                            CDN77.purge(files_to_purge_list, request_number)


from django.contrib.auth.hashers import make_password, check_password
from django.conf import settings
import math


def _pepper(password):
    pepper1 = settings.BULB_PEPPER_1
    pepper2 = settings.BULB_PEPPER_2
    password_length = len(password)
    splitted_password = list(password)
    splitted_password.insert(math.floor(password_length / 2), pepper1)
    splitted_password.append(pepper2)
    pepper_password = ''.join(splitted_password)
    return pepper_password


def _hash_password(raw_password):
    hashed_password = make_password(_pepper(raw_password), hasher="bcrypt_sha256")
    return hashed_password


def _check_password(raw_password, encoded_password):
    if check_password(_pepper(raw_password), encoded_password):
        return True
    else:
        return False
